/* (c) Joerg Schulenburg Jun2000-2007

 this is part of spin-package

 it is for division of matrix multiplication 
 in parallelizable block matrix multiplication (see block.h)

 */
#include "block.h"

int mpi_n=1, mpi_i=0;   /* number of MPI nodes, no. of this node */
int pt_n=1;             /* number of threads per node 1..B_NUM */
int mpi_n_bits=1, pt_n_bits=1; /* bits needed to store mpi_i and pt_i, MPI */
/* ToDo: only 32 bit? -> ULLong */
unsigned long node_ofs, node_n1; /* ofset and length of subspace this node */
/* tables for each thread accessed via [mpi_i*pt_n+pt_i] */
unsigned long *b_ofs=NULL;	/* start index 0..n1-1 of vector block */
unsigned long *b_len=NULL;	/* vector block length */
/* ToDo: move b_hnz to h_hnz h_file.c!? where h_ is defined */
unsigned long b_hnz[B_NUM];	/* block NonZero H-entries */
unsigned long b_snz[B_NUM];	/* stored hnz block entries */
unsigned long b_sn1[B_NUM];	/* stored hnz block lines */
unsigned long len_full_block;   /* number of elements per block */

/* ToDo: move hbuf to h_file.c!? where thxy is defined */
struct thxy b_hbuf[B_NUM];      /* buffer for faster r/w-access */ 
int         b_hbl[B_NUM];       /* num valid data in buffer hbuf */  

/* release mem to satisfy memleak checker */
void b_free(){
   if (b_ofs) free(b_ofs); b_ofs=NULL;
   if (b_len) free(b_len); b_len=NULL;
}

/* initialization of thread blocks for n1/mpi_n elements on each node,
 * n1 is whole matrix size, called by every node, only thread0
 */
int b_ini(unsigned long n1,int vvv){
  int i, num_full_blocks;
  unsigned long len_rest_block, ofs, len;
  
  if (!b_ofs) b_ofs=(unsigned long *)malloc(pt_n*mpi_n*sizeof(unsigned long));
  if (!b_ofs) ERR(9011);

  if (!b_len) b_len=(unsigned long *)malloc(pt_n*mpi_n*sizeof(unsigned long));
  if (!b_len) ERR(9011);
          
  /* pt_n is numthreads per node */
  for (i=0;i<pt_n*mpi_n;i++) {  /* initialize all arrays (1..pt_n*mpi_n)-1 */
    b_ofs[i]=0;
    b_len[i]=0;
  }
  for (i=0;i<B_NUM;i++) {  /* initialize all arrays (1..pt_n..B_NUM)-1 */
    b_sn1[i]=0; /* belongs to h_file!? */
    b_hnz[i]=0;
    b_snz[i]=0;
    b_hbl[i]=0;
  }
  /* block length depends from number of mpi nodes and threads per node */
  len_full_block=(n1+pt_n*mpi_n-1)/(pt_n*mpi_n); /* #loops/thread 0..n1 */
#if defined(HAVE_MMAP) && defined(CFG_PAGESIZE)
  /* mmap offset must be a multiple of PAGESIZE! fill up block length */
  if (len_full_block>0)
      len_full_block=((len_full_block-1)/CFG_PAGESIZE+1)*CFG_PAGESIZE;
#endif /* CFG_PAGESIZE */

  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (!mpi_i)
    printf("\n b_ini: mpi_n= %d pt_n= %d n1= %lu len_full_blk= %lu\n",
       mpi_n, pt_n, n1, len_full_block);
  /* calculate num of full size blocks and length rest block */
  num_full_blocks= n1 / len_full_block;     /* range: 0..(pt_n*mpi_n-1) */
  len_rest_block=  n1 - len_full_block * num_full_blocks;  /* 0..(n1-1) */

  if (!mpi_i) {
    printf(" blocks:  num   * length  + rest = %3d * %8lu + %8lu = n1\n",
       num_full_blocks, len_full_block, len_rest_block);
  }
  node_ofs=0; node_n1=0;
  MPI_Barrier(MPI_COMM_WORLD); /* debug */

  /* setup blocks (offsets and lengths) */
  ofs=0; len=n1;
  for (i=0; i<mpi_n*pt_n; i++) {
    b_ofs[i]=ofs;
    b_len[i]=(len<len_full_block)?len:len_full_block;
    if (i/pt_n==mpi_i) node_n1+=b_len[i]; /* length l1-part this node */
    ofs+=b_len[i];
    len-=b_len[i];
    if (b_len[i] && (!mpi_i) && (i==4 && i<=mpi_n*pt_n-8))
      printf(" MPI_DBG: ...\n");
    if (b_len[i] && (!mpi_i) && (i< 4 || i> mpi_n*pt_n-8))
      printf(" MPI_DBG: node=%3d blk=%3d b_ofs= %8lu b_len= %8lu\n",
              i/pt_n, i%pt_n, b_ofs[i], b_len[i]);
  }
  node_ofs=b_ofs[mpi_i*pt_n];  /* node offset, do we need it? */

  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  printf(    " MPI_DBG: node=%3d   node_ofs= %8lu node_len= %8lu\n", 
         mpi_i, node_ofs, node_n1);
  MPI_Barrier(MPI_COMM_WORLD); /* debug */

  if (vvv>1 && mpi_i==0) /* reduce the output */
    for (i=0; i<pt_n*mpi_n; i++) { 
      if (i==0 || b_len[i]!=b_len[i-1])
      printf(" node %2d block %2d v_ofs %8lu v_len %8lu\n", 
             i/pt_n, i%pt_n, b_ofs[i], b_len[i]);
    }
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  return 0;
}
