/*
  linux: mpicc -O2 -g -Wall ../utils/mpi_speed.c
  P4-1.3GHz-linux 15.3us      476MB/s    (2)
  ES45-Tru64-5.1B  2.6us 5%   820MB/s    (2)
                   2.79us    1130MB/s 2% (4)
      overload-2x 26.0us 4%   7.2MB/s 0.2%(8) 
  GS1280-32:       3.1us 4%   715MB/s 4% (2) (slow 32CPUs/RAD)
                   3.9us 3%  1960MB/s(512kB)
                             1560MB/s(1MB) 3% (8)
                   6.4us 4%  3930MB/s(32kB)
                             4040MB/s(64kB,128kB)
                             4160MB/s(256kB)
                             3500MB/s(512kB)
                             2500MB/s(1MB) 3% (32) 
  Altix330 numa_alloc_local MPI_via_loopback ToDo: interconnect-topology=star
           MPT-MPI-1.2 (insmod xpmem) MPI_USE_XPMEM=1 MPI_XPMEM_VERBOSE=1 ?
            2 2.09us 1.8GB/s(1M) 1.2GB/s 1.1GB/s(256K) 1node  numa_alloc_local
            2 1.97us 8G(1M) 17G 15G 13G 12G 8G 5G 3G 2G 1G 300M ... no div psr1,11
            4        1.6GB/s(1M) 1.7GB/s 1.6GB/s(256K) 4nodes
            8 3.58us 2.5GB/s(1M) 2.6GB/s 2.7GB/s(256K) 4nodes
           12 4.05us 3.8GB/s(1M) 3.4GB/s 3.4GB/s(256K) 6nodes numa_alloc_local
           12 4.05us 4.1GB/s(1M) 4.5GB/s  14GB/s 34GB/s(128K) 6nodes malloc lo
           12 4.05us 4.3GB/s(1M) 5.0GB/s  12GB/s 32GB/s(128K) 6nodes malloc xpmem
  2DualAMD265:      23us 4%   660MB/s 4% (last!)   baldur
  4Xeon3GHz_tina   3.3us 4%   307MB/s 4% (4) Dec07 node004
    2*1G via .1G   160us 5%    15MB/s 3% (2) Dec07 node4+node012
                   191us 4%    28MB/s 3% (4) Dec07 node4+node012+...
                   206us 4%    63MB/s 3% (8) Dec07 node4+node012+...

*/
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h> 
#include <time.h>

#define BUFSIZE (1024*1024)
int mpi_n, mpi_i;
char *buf1, *buf2;
static int header=0;

int testrun(unsigned long len){
  unsigned long i1, i2, i3;
  int node1, node2; time_t t1, t2;
  MPI_Status status;
  if (!header) { header=1;
    if (!mpi_i)
    printf("#    msgsize[B] loops[2^x] time[s]"
           " latency[us] SumBW[MB/s] error\n"); /* BW=bandwidth*/
  }
  i3=8;
  for (i1=1l<<i3;i1;i1<<=1,i3++){
    time(&t1);
    for (i2=0;i2<i1;i2++) {
      node1=(mpi_i-(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
      node2=(mpi_i+(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
      MPI_Sendrecv(buf1,len,MPI_BYTE,node1,0,
                   buf2,len,MPI_BYTE,node2,0,MPI_COMM_WORLD,&status);
    }
    time(&t2); t2-=t1;
    if (t2>1 && !mpi_i)
    printf("SumNodes %8lu %6lu %8d %9.2f  %8.2f  %5.1f%%\n",
           len, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n/t2, 100./t2);
    if (t2>20) break;
  }
  if (!mpi_i)
  printf("SumNodes %8lu %6lu %8d %9.2f  %8.2f  %5.1f%% <= final\n",
           len, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n/t2, 100./t2);
  return 0;
}


int main(int argc, char *argv[]) {
  unsigned long i1;
  setvbuf(stdout,NULL,_IONBF,0);
  if (MPI_Init(&argc, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  { int version, subversion;
  MPI_Get_version(&version,&subversion);
  if (!mpi_i)
   printf(" using MPI %d.%d with %d nodes\n", version, subversion, mpi_n);
  }
  buf1=(char *)malloc(BUFSIZE); /* replace by numa_alloc_local */
  buf2=(char *)malloc(BUFSIZE);
  if ((!buf1) || (!buf2)) exit(1);
  for (i1=0;i1<BUFSIZE;i1++) buf1[i1]=buf2[i1]=(char)i1;
  MPI_Barrier(MPI_COMM_WORLD);
 
  if (!mpi_i) printf("Measure speed of MPI_sendrecv transactions:\n");
  testrun(BUFSIZE);
  testrun(BUFSIZE/2);
  testrun(BUFSIZE/4);
  testrun(BUFSIZE/8);
  testrun(BUFSIZE/16);
  testrun(1);
  testrun(2);

  free(buf2);
  free(buf1);  
  MPI_Finalize();
  return 0;
}
